from skitai.rpc import cluster_manager
from aquests.dbapi import asynpsycopg2, synsqlite3, asynredis, asynmongo
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB

class ClusterManager (cluster_manager.ClusterManager):
    backend_keep_alive = 1200
    backend = True
    
    def __init__ (self, name, cluster, dbtype = DB_PGSQL, access = [], logger = None):
        self.dbtype = dbtype
        self._cache = []
        cluster_manager.ClusterManager.__init__ (self, name, cluster, 0, access, logger)
            
    def match (self, request):
        return False # not serverd by url
    
    def create_asyncon (self, member):        
        if self.dbtype == DB_SQLITE3:
            asyncon = synsqlite3.SynConnect (member, None, self.lock, self.logger)
            nodeid = member
            self._cache.append (((member, 0), "", ("", "")))
        
        else:
            if member.find ("@") != -1:
                auth, netloc = self.parse_member (member)
                try:
                    server, db = netloc.split ("/", 1)
                except ValueError:
                    server, db = netloc, ""
                    
            else:                
                db, user, passwd = "", "", ""
                args = member.split ("/", 3)
                if len (args) == 4:     server, db, user, passwd = args
                elif len (args) == 3:     server, db, user = args
                elif len (args) == 2:     server, db = args        
                else: server = args [0]
                auth = (user, passwd)
                
            try: 
                host, port = server.split (":", 1)
                server = (host, int (port))
            except ValueError: 
                server = (server, 5432)
            
            if self.dbtype == DB_PGSQL:
                conn_class = asynpsycopg2.AsynConnect
            elif self.dbtype == DB_REDIS:
                conn_class = asynredis.AsynConnect
            elif self.dbtype == DB_MONGODB:
                conn_class = asynmongo.AsynConnect    
            else:
                raise TypeError ("Unknown DB type: %s" % self.dbtype)
            
            asyncon = conn_class (server, (db, auth), self.lock, self.logger)    
            self.backend and asyncon.set_backend (self.backend_keep_alive)            
            nodeid = server
            self._cache.append ((server, db, auth))
                
        return nodeid, asyncon # nodeid, asyncon
    
    def get_endpoints (self):
        return make_endpoints (self.dbtype, self._cache)
        

def make_endpoints (dbtype, from_list):
        import sqlite3
        import psycopg2
        import redis
        import pymongo
        
        endpoints = []        
        for server, db, auth in from_list:
            user, password = "", ""
            if auth:
                if len (auth) == 2:
                    user, password = auth
                else:
                    user = auth [0]    
            if isinstance (server, str):
                try: 
                    host, port = server.split (":", 1)
                except ValueError:
                    host, port = server, None
            else:
                 host, port = server
            
            kargs = {}
            if port: kargs ["port"] = port
            if password: kargs ["password"] = password
            
            if dbtype == DB_SQLITE3:
                conn = sqlite3.connect (host)
            elif dbtype == DB_PGSQL:
                if user: kargs ["user"] = user
                conn = psycopg2.connect (host = host, database = db, **kargs)
            elif dbtype == DB_REDIS:
                conn = redis.Redis (host = host, port = port, db = db)
            elif dbtype == DB_MONGODB:
                if user: kargs ["username"] = user
                conn = pymongo.MongoClient (host = host, **kargs)
            endpoints.append (conn)
        return endpoints

       
              


import os, sys
import time
import tempfile
from hmac import new as hmac
from hashlib import sha1
import json
import xmlrpc.client as xmlrpclib
import base64
import pickle
import random
import threading
from rs4 import pathtool, logger, jwt
from rs4.producers import simple_producer, file_producer
from aquests.athreads import trigger
from aquests.protocols.smtp import composer
from aquests.protocols.http import http_date, util
import inspect
from skitai import __version__, WS_EVT_OPEN, WS_EVT_CLOSE, WS_EVT_INIT, NAME
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from skitai import lifetime
from skitai.handlers import api_access_handler, vhost_handler
from .rpc import cluster_manager, cluster_dist_call
from .dbi import cluster_manager as dcluster_manager, cluster_dist_call as dcluster_dist_call
from . import server_info
from . import http_response
from .wastuff.promise import Promise
from .wastuff.triple_logger import Logger
from .wastuff import django_adaptor
from .wastuff.api import DateEncoder
from rs4.attrdict import AttrDict
if os.environ.get ("SKITAI_ENV") == "PYTEST":
    from threading import RLock
else:    
    from multiprocessing import RLock

if os.name == "nt":
    TEMP_DIR =  os.path.join (tempfile.gettempdir(), "skitai-gentemp")
else:
    TEMP_DIR = "/var/tmp/skitai-gentemp"        
pathtool.mkdir (TEMP_DIR)

class JWTUser:
    def __init__ (self, claims):
        self.__claims = claims
    
    @property
    def name (self):
        return self.__claims ["username"]
        
    def __getattr__ (self, attr):
        return self.__claims.get (attr)
            
    def __str__ (self):
        return self.name

def is_main_thread ():    
    return isinstance (threading.currentThread (), threading._MainThread)

class Command:
    def __init__ (self, name, callback):
        self.name = name
        self.callback = callback
    
    def __call__ (self, *args, **kargs):
        return self.callback (self.name, args, kargs)
    
    def lb (self, *args, **kargs):
        return self.callback (self.name + ".lb", args, kargs)
    
    def map (self, *args, **kargs):
        return self.callback (self.name + ".map", args, kargs)


class WAS:
    version = __version__    
    objects = {}    
    _luwatcher = None
    
    lock = RLock ()
    init_time = time.time ()    
    
    METHODS = [
        "options", "trace", "upload",
        "get", "delete", "post", "put", "patch",        
        "rpc", "xmlrpc", "jsonrpc", "grpc", "ws", "wss", 
        "db", "postgresql", "sqlite3", "redis", "mongodb", "backend",        
    ]
    DEFAULT_REQUEST_TYPE = ("application/json", "application/json")
    
    def __init__ (self):
        for method in self.METHODS:
            setattr (self, method, Command (method, self._call))            
    
    # application friendly methods -----------------------------------------
    @classmethod
    def register (cls, name, obj):
        if hasattr (cls, name):
            raise AttributeError ("server object `%s` is already exists" % name)
        cls.objects [name] = obj
        setattr (cls, name, obj)
    
    @classmethod
    def unregister (cls, name):
        del cls.objects [name]
        return delattr (cls, name)
        
    @classmethod
    def add_handler (cls, back, handler, *args, **karg):
        h = handler (cls, *args, **karg)
        if hasattr (cls, "httpserver"):
            cls.httpserver.install_handler (h, back)
        return h
                        
    @classmethod
    def add_cluster (cls, clustertype, clustername, clusterlist, ssl = 0, access = []):
        if clustertype and clustertype [0] == "*":
            clustertype = clustertype [1:]
        ssl = 0
        if ssl in (1, True, "1", "yes") or clustertype in ("https", "wss", "grpcs", "rpcs"):
            ssl = 1
        if type (clusterlist) is str:
            clusterlist = [clusterlist]    

        if clustertype and "*" + clustertype in (DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB):
            cluster = dcluster_manager.ClusterManager (clustername, clusterlist, "*" + clustertype, access, cls.logger.get ("server"))
            cls.clusters_for_distcall [clustername] = dcluster_dist_call.ClusterDistCallCreator (cluster, cls.logger.get ("server"))
        else:
            cluster = cluster_manager.ClusterManager (clustername, clusterlist, ssl, access, cls.logger.get ("server"))
            cls.clusters_for_distcall [clustername] = cluster_dist_call.ClusterDistCallCreator (cluster, cls.logger.get ("server"), cls.cachefs)
        cls.clusters [clustername] = cluster
    
    @property
    def timestamp (self):
        return int (time.time () * 1000) 
    
    @property
    def uniqid (self):
        return "{}{}".format (self.timestamp, self.gentemp () [-7:])
        
    def __dir__ (self):
        return self.objects.keys ()
    
    def __str__ (self):
        return "skitai was for {}".format (threading.currentThread ())
                    
    def __detect_cluster (self, clustername):
        try: 
            clustername, uri = clustername.split ("/", 1)
        except ValueError:
            clustername, uri = clustername, ""
        if clustername [0] == "@":
            clustername = clustername [1:]
        
        try: 
            return self.clusters_for_distcall ["{}:{}".format (clustername, self.app.app_name)], "/" + uri            
        except (KeyError, AttributeError):
            return self.clusters_for_distcall [clustername], "/" + uri
        
    def in__dict__ (self, name):
        return name in self.__dict__
    
    def _clone (self, disable_aquests = False):
        new_was = self.__class__ ()
        for k, v in self.__dict__.items ():
            setattr (new_was, k, v)    
        if disable_aquests:
            for method in self.METHODS:
                delattr (new_was, method)
        return new_was
    
    # mehiods remap ------------------------------------------------
    def __getattr__ (self, name):
        if self.in__dict__ ("app"): # atila app
            attr = self.app.create_on_demand (self, name)
            if attr:
                setattr (self, name, attr)
                return attr
        
        try:
            return self.objects [name]
        except KeyError:    
            raise AttributeError ("'was' hasn't attribute '%s'" % name)    
        
    def _call (self, method, args, karg):
        # was.db,         was.get,             was.post,            was.put, ...
        # was.db.lb,     was.get.lb,        was.post.lb,    was.put.lb, ...
        # was.db.map,    was.get.map,    was.post.map,    was.put.map, ...
        uri = None
        if args:        uri = args [0]
        elif karg:    uri = karg.get ("uri", "")
        if not uri:    raise AssertionError ("Missing param uri or cluster name")

        try: 
            command, fn = method.split (".")
        except ValueError: 
            command = method
            if uri [0] == "@": 
                fn = "lb"
            else:
                fn = (command in {"db", "postgresql", "sqlite3", "redis", "mongodb", "backend"} and "db" or "rest")

        if fn == "map" and not hasattr (self, "threads"):
            raise AttributeError ("Cannot use Map-Reduce with Single Thread")
                
        if command == "db":
            return getattr (self, "_d" + fn) (*args, **karg)            
        elif command in ("postgresql", "sqlite3", "redis", "mongodb", "backend"):
            return getattr (self, "_a" + fn) ("*" + command, *args, **karg)                    
        else:    
            return getattr (self, "_" + fn) (command, *args, **karg)
    
    # TXN -----------------------------------------------
    
    def txnid (self):
        return "%s/%s" % (self.request.gtxid, self.request.ltxid)
    
    def rebuild_header (self, header, method, data = None, internal = True):
        nheader = util.normheader (header)
        if method in {"get", "delete", "post", "put", "patch", "upload"}:
            try:
                default_request_type = self.app.config.get ("default_request_type")
            except AttributeError:
                default_request_type = None
            if not default_request_type:    
                default_request_type = self.DEFAULT_REQUEST_TYPE            
            util.set_content_types (nheader, data, default_request_type)
            
        if internal:            
            nheader ["X-Gtxn-Id"] = self.request.get_gtxid ()
            nheader ["X-Ltxn-Id"] = self.request.get_ltxid (1)
        else:
            nheader ["X-Requested-With"] = NAME
        return nheader
    
    # async requests -----------------------------------------------
    def assert_params (self, callback, mapreduce = False):
        if is_main_thread () and not callback:
            raise RuntimeError ('Should have callback in Main thread')
        if mapreduce and callback: 
            raise RuntimeError ("Cannot use callback with Map-Reduce")
        
    def _create_rest_call (self, cluster, *args, **kargs):
        return cluster.Server (*args, **kargs)
            
    def _rest (self, method, uri, data = None, auth = None, headers = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
        return self._create_rest_call (self.clusters_for_distcall ["__socketpool__"], uri, data, method, self.rebuild_header (headers, method, data, False), auth, meta, use_cache, False, filter, callback, timeout, caller)
    
    def _crest (self, mapreduce = False, method = None, uri = None, data = None, auth = None, headers = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
        cluster, uri = self.__detect_cluster (uri)
        return self._create_rest_call (cluster, uri, data, method, self.rebuild_header (headers, method, data), auth, meta, use_cache, mapreduce, filter, callback, timeout, caller)
                
    def _lb (self, *args, **karg):
        return self._crest (False, *args, **karg)
        
    def _map (self, *args, **karg):
        return self._crest (True, *args, **karg)
    
    def _bind_sqlphile (self, dbo):
        try:
            app_sqlphile = self.sql
        except AttributeError:
            return dbo    
        return app_sqlphile.new (dbo)
    
    def _create_dbo (self, cluster, *args, **kargs):
        return cluster.Server (*args, **kargs)
            
    def _ddb (self, server, dbname = "", auth = None, dbtype = DB_PGSQL, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
        dbo = self._create_dbo (self.clusters_for_distcall ["__dbpool__"], server, dbname, auth, dbtype, meta, use_cache, False, filter, callback, timeout, caller)
        if dbtype in (DB_PGSQL, DB_SQLITE3):
            return self._bind_sqlphile (dbo)
        return dbo
    
    def _cddb (self, mapreduce = False, clustername = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
        cluster = self.__detect_cluster (clustername) [0]
        dbo = self._create_dbo (cluster, None, None, None, None, meta, use_cache, mapreduce, filter, callback, timeout, caller)
        if cluster.cluster.dbtype in (DB_PGSQL, DB_SQLITE3):
            return self._bind_sqlphile (dbo)
        return dbo    
    
    def _dlb (self, *args, **karg):
        return self._cddb (False, *args, **karg)
    
    def _dmap (self, *args, **karg):
        return self._cddb (True, *args, **karg)
    
    def _adb (self, dbtype, server, dbname = "", auth = None, meta = None, use_cache = True, filter = None, callback = None, timeout = 10, caller = None):
        return self._ddb (server, dbname, auth, dbtype, meta, use_cache, filter, callback, timeout, caller)
    
    def _alb (self, dbtype, *args, **karg):
        return self._cddb (False, *args, **karg)
    
    def _amap (self, dbtype, *args, **karg):
        return self._cddb (True, *args, **karg)
    
    # system functions ----------------------------------------------
        
    def log (self, msg, category = "info", at = "app"):
        self.logger (at, msg, "%s:%s" % (category, self.txnid ()))
        
    def traceback (self, id = "", at = "app"):
        if not id:
            id = self.txnid ()
        self.logger.trace (at, id)
    
    @property
    def tempfile (self):
        return self.gentemp () 
    
    def gentemp (self):
        return os.path.join (TEMP_DIR, next (tempfile._get_candidate_names()))
    
    # -- only allpy to current worker process
    def status (self, flt = None, fancy = True):        
        return server_info.make (self, flt, fancy)
    
    def restart (self, timeout = 0):
        lifetime.shutdown (3, timeout)
    
    def shutdown (self, timeout = 0):
        lifetime.shutdown (0, timeout)
    
    # URL builders -------------------------------------------------
    def urlfor (self, thing, *args, **karg):
        # override with resource default args
        if not isinstance (thing, str) or thing.startswith ("/") or thing.find (":") == -1:
            return self.app.urlfor (thing, *args, **karg)        
        return self.apps.urlfor (thing, *args, **karg)    
    ab = urlfor
    
    def partial (self, thing, **karg):
        # override with current args
        karg ["__defaults__"] = self.request.args
        return self.ab (thing, **karg)
    
    def baseurl (self, thing):
        # resource path info without parameters
        return self.ab (thing, __resource_path_only__ = True)
    basepath = baseurl
    
    # response helpers --------------------------------------------
        
    def render (self, template_file, _do_not_use_this_variable_name_ = {}, **karg):
        return self.app.render (self, template_file, _do_not_use_this_variable_name_, **karg)
    
    REDIRECT_TEMPLATE =  (
        "<html><head><title>%s</title></head>"
        "<body><h1>%s</h1>"
        "This document may be found " 
        '<a HREF="%s">here</a></body></html>'
    )
    def redirect (self, url, status = "302 Object Moved", body = None, headers = None):
        redirect_headers = [
            ("Location", url), 
            ("Cache-Control", "max-age=0"), 
            ("Expires", http_date.build_http_date (time.time ()))
        ]
        if type (headers) is list:
            redirect_headers += headers
        if not body:
            body = self.REDIRECT_TEMPLATE % (status, status, url)            
        return self.response (status, body, redirect_headers)
    
    def promise (self, handler, **karg):
        self.response.set_streaming ()
        return Promise (self, handler, **karg)
    
    def email (self, subject, snd, rcpt):
        if composer.Composer.SAVE_PATH is None:
            composer.Composer.SAVE_PATH = os.path.join ("/var/tmp/skitai", "smtpda", "spool")
            pathtool.mkdir (composer.Composer.SAVE_PATH)
        return composer.Composer (subject, snd, rcpt)
    
    # event -------------------------------------------------
    
    def broadcast (self, event, *args, **kargs):
        return self.apps.bus.emit (event, self, *args, **kargs)
    
    def setlu (self, name, *args, **karg):
        self._luwatcher.set (name, time.time (), karg.get ("x_ignore", False))
        self.broadcast (name, *args, **karg)            
        
    def getlu (self, *names):
        mtimes = []
        for name in names:
            mtime = self._luwatcher.get (name, self.init_time)
            mtimes.append (mtime)
        return max (mtimes)
    
    # JWT token --------------------------------------------------
    def mkjwt (self, claim, alg = "HS256"):
        assert "exp" in claim, "exp claim required"            
        return jwt.gen_token (self.app.salt, claim, alg)
        
    def dejwt (self, token = None):
        if not token:
            token_ = self.request.get_header ("authorization")
            if not token_ or token_ [:7].lower () != "bearer ":
                return
            token = token_ [7:]
            
        try: 
            claims = jwt.get_claim (self.app.salt, token)
        except (TypeError, ValueError): 
            return {"err": "invalid token"}
        if claims is None:
            return {"err": "invalid signature"}
        now = time.time ()
        if claims ["exp"] < now:
            return {"err": "token expired"}
        if "nbf" in claims and claims ["nbf"] > now:
            return {"err": "token not activated yet"}
        if "username" in claims:
            self.request.user = JWTUser (claims)
        self.request.JWT = claims    
        return claims
        
    # simple/session token  ----------------------------------------
    def _unserialize_token (self, string):
        def adjust_padding (s):
            paddings = 4 - (len (s) % 4)
            if paddings != 4:
                s += ("=" * paddings)
            return s
        
        string = string.replace (" ", "+")
        try:
            base64_hash, data = string.split('?', 1)
        except ValueError:
            return    
        client_hash = base64.b64decode(adjust_padding (base64_hash))
        data = base64.b64decode(adjust_padding (data))
        mac = hmac (self.app.salt, None, sha1)        
        mac.update (data)
        if client_hash != mac.digest():
            return
        return pickle.loads (data)
    
    def mkott (self, obj, timeout = 1200, session_key = None):
        wrapper = {
            'object': obj,
            'timeout': time.time () + timeout
        }
        if session_key:
            token = hex (random.getrandbits (64))
            tokey = '_{}_token'.format (session_key)
            wrapper ['_session_token'] = (tokey, token)
            self.session [tokey] = token
            
        data = pickle.dumps (wrapper, 1)
        mac = hmac (self.app.salt, None, sha1)
        mac.update (data)
        return (base64.b64encode (mac.digest()).strip().rstrip (b'=') + b"?" + base64.b64encode (data).strip ().rstrip (b'=')).decode ("utf8")
    
    def deott (self, string):
        wrapper = self._unserialize_token (string)
        if not wrapper:
            return 
        
        # validation with session
        tokey = None
        has_error = False
        if wrapper ['timeout']  < time.time ():
            has_error = True

        if not has_error:
            session_token = wrapper.get ('_session_token')
            if session_token:
                # verify with session                
                tokey, token = session_token    
                if token != self.session.get (tokey):
                    has_error = True
                    
        if has_error:
            if tokey:
                del self.session [tokey]
            return
        
        obj = wrapper ['object']
        return obj
    
    def rvott (self, string):
        # revoke token
        wrapper = self._unserialize_token (string)
        session_token = wrapper.get ('_session_token')
        if not session_token:
            return
        tokey, token = session_token
        if not self.session.get (tokey):
            return
        del self.session [tokey]
    
    mktoken = token = mkott
    rmtoken = rvott
    detoken = deott
        
    # CSRF token ------------------------------------------------------    
    CSRF_NAME = "_csrf_token"
    @property
    def csrf_token (self):
        if self.CSRF_NAME not in self.session:
            self.session [self.CSRF_NAME] = hex (random.getrandbits (64))
        return self.session [self.CSRF_NAME]

    @property
    def csrf_token_input (self):
        return '<input type="hidden" name="{}" value="{}">'.format (self.CSRF_NAME, self.csrf_token)
    
    def csrf_verify (self, keep = False):
        if not self.request.args.get (self.CSRF_NAME):
            return False
        token = self.request.args [self.CSRF_NAME]
        if self.csrf_token == token:
            if not keep:
                del self.session [self.CSRF_NAME]
            return True
        return False
    
    # proxy & adaptor  -----------------------------------------------
    @property
    def sql (self):
        return self.app.sqlphile
    
    @property
    def django (self):
        if hasattr (self.request, "django"):
            return self.request.django
        self.request.django = django_adaptor.request (self)
        return self.request.django
    
    # websocket methods for generic WSGI containers -----------------------------
    
    def wsconfig (self, spec, timeout = 60, encoding = "text"):
        self.env ["websocket.config"] = (spec, timeout, encoding)
        return ""
        
    def wsinit (self):
        return self.env.get ('websocket.event') == WS_EVT_INIT
    
    def wsopened (self):
        return self.env.get ('websocket.event') == WS_EVT_OPEN
    
    def wsclosed (self):
        return self.env.get ('websocket.event') == WS_EVT_CLOSE
    
    def wshasevent (self):
        return self.env.get ('websocket.event')
    
    def wsclient (self):
        return self.env.get ('websocket.client')    
            
    # will be deprecated --------------------------------------------------    
    
    def togrpc (self, obj):
        return obj.SerializeToString ()
    
    def fromgrpc (self, message, obj):
        return message.ParseFromString (obj)
        
    def tojson (self, obj):
        return json.dumps (obj, cls = DateEncoder)
        
    def toxml (self, obj):
        return xmlrpclib.dumps (obj, methodresponse = False, allow_none = True, encoding = "utf8")    
    
    def fromjson (self, obj):
        if type (obj) is bytes:
            obj = obj.decode ('utf8')
        return json.loads (obj)
    
    def fromxml (self, obj, use_datetime = 0):
        return xmlrpclib.loads (obj)
    
    def fstream (self, path, mimetype = 'application/octet-stream'):    
        self.response.set_header ('Content-Type',  mimetype)
        self.response.set_header ('Content-Length', str (os.path.getsize (path)))    
        return file_producer (open (path, "rb"))
            
    def jstream (self, obj, key = None):
        self.response.set_header ("Content-Type", "application/json")
        if key:
            # for single skeleton data is not dict
            return self.tojson ({key: obj})
        else:
            return self.tojson (obj)        
    
    def xstream (self, obj, use_datetime = 0):            
        self.response.set_header ("Content-Type", "text/xml")
        return self.toxml (obj, use_datetime)
    
    def gstream (self, obj):
        self.response.set_header ("Content-Type", "application/grpc")
        return self.togrpc (obj)
    
    def render_ei (self, exc_info, format = 0):
        return http_response.catch (format, exc_info)
        


# Web Socket message dump and load related codes are based on websocket-server 0.4
# by Johan Hanssen Seferidis
# https://pypi.python.org/pypi/websocket-server
#
# 2016. 1. 16 Modified by Hans Roh

from . import wsgi_handler
from hashlib import sha1
from base64 import b64encode
from skitai.http_response import catch
from aquests.protocols.http import http_util
from skitai import version_info, was as the_was
import threading		
from .websocket import specs
from .websocket import servers
import time
import skitai
import inspect

class Handler (wsgi_handler.Handler):
	def match (self, request):
		upgrade = request.get_header ("upgrade")
		return upgrade and upgrade.lower ().startswith ("websocket") and request.version == "1.1" and request.command == "get"
	
	def close (self):
		servers.websocket_servers.close ()
	
	GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'.encode ()
	def calculate_response_key (self, key):		
		hash = sha1(key.encode() + self.GUID)
		response_key = b64encode(hash.digest()).strip()
		return response_key.decode('ASCII')
		
	def handle_request (self, request):
		def donot_response (self, *args, **kargs):
			def push (thing):
				raise AssertionError ("Websocket can't use start_response ()")
			return push
		
		origin = request.get_header ("origin")
		host = request.get_header ("host")
		protocol = request.get_header ("sec-websocket-protocol", 'unknown')
		securekey = request.get_header ("sec-websocket-key")
		
		if not origin or not host or not securekey:
			return request.response.error (400)
		
		path, params, query, fragment = request.split_uri ()
		has_route = self.apps.has_route (path)
		if type (has_route) is int:
			return request.response.error (404)
		
		apph = self.apps.get_app (path)		
		app = apph.get_callable()
		is_atila = skitai.HAS_ATILA and isinstance (app, skitai.HAS_ATILA)
		
		if is_atila:
			# safari does not support Authorization
			if request.get_header ("authorization") and not app.is_authorized (request, app.authenticate):
				return request.response.error (401)
			if not app.is_allowed_origin (request, app.access_control_allow_origin):
				return request.response.error (403)
				
		env = self.build_environ (request, apph)
		was = the_was._get ()
		was.request = request
		env ["skitai.was"] = was
		env ["websocket.event"] = skitai.WS_EVT_INIT

		message_encoding = skitai.WS_MSG_DEFAULT				
		if not is_atila:	# not Skitao-Atila				
			apph (env, donot_response)
			wsconfig = env.get ("websocket.config", ())			
			if len (wsconfig) == 3:
				design_spec, keep_alive, varnames = wsconfig
				if type (varnames) not in (list, tuple):
					varnames = (varnames,)
			else:
				raise AssertionError ("You should config (design_spec, keep_alive, var_names) where env has key 'skitai.websocket.config'")
				
		else:	
			current_app, method, kargs, options, resp_code = apph.get_callable().get_method (env ["PATH_INFO"], request)
			if resp_code:
				return request.response.error (resp_code)
				
			request.routed = current_app.get_routed (method)
			request.routable = options	
			
			wsfunc = request.routed
			if is_atila:
				fspec = app.get_function_spec (wsfunc) or inspect.getfullargspec (wsfunc)				
			else:
				fspec = inspect.getfullargspec (wsfunc)
			
			savedqs = env.get ('QUERY_STRING', '')
			current_args = {}
			defaults = 0
			if savedqs:
				current_args = http_util.crack_query (env ['QUERY_STRING'])
			if fspec.defaults:
				defaults = len (fspec.defaults)
			varnames = fspec.args [1:]			
			temporary_args = "&".join ([arg + "=" for arg in varnames [:len (varnames) - defaults] if current_args.get (arg) is None])			
			if temporary_args:
				if savedqs:
					env ['QUERY_STRING'] = savedqs + "&" + temporary_args
				else:
					env ['QUERY_STRING'] = temporary_args
			
			apph (env, donot_response)
			wsconfig = env.get ("websocket.config")			
			if not wsconfig:
				raise AssertionError ("You should config (design_spec, keep_alive, [data_encoding]) where env has key 'was.wsconfig ()'")
			
			if not savedqs and "QUERY_STRING" in env:
				del env ["QUERY_STRING"]
			else:	
				env ["QUERY_STRING"] = savedqs
			
			keep_alive = 60
			if len (wsconfig) == 3:
				design_spec, keep_alive, message_encoding = wsconfig					
			elif len (wsconfig) == 2:
				design_spec, keep_alive = wsconfig
			elif len (wsconfig) == 1:
				design_spec = wsconfig [0]	
				
		del env ["websocket.event"]
		del env ["websocket.config"]
		assert design_spec in (1,5,6), "design_spec  should be one of (WS_SIMPLE, WS_GROUPCHAT, WS_THREADSAFE)"			
		headers = [
			("Sec-WebSocket-Accept", self.calculate_response_key (securekey)),
			("Upgrade", "Websocket"),
			("Connection", "Upgrade"),
		    ("WebSocket-Protocol", protocol),
	    	("WebSocket-Location", "ws://" + host + path)
		]
		request.response ("101 Web Socket Protocol Handshake", headers = headers)
		
		if design_spec in (skitai.WS_SIMPLE, skitai.WS_THREADSAFE):
			varnames = varnames [:1]
			# Like AJAX, simple request of client, simple response data
			# the simplest version of stateless HTTP protocol using basic skitai thread pool
			ws_calss = design_spec == skitai.WS_SIMPLE and specs.WebSocket1 or specs.WebSocket6			
			ws = ws_calss (self, request, apph, env, varnames, message_encoding)
			self.channel_config (request, ws, keep_alive)
			env ["websocket"] = ws
			if is_atila: env ["websocket.handler"] = (current_app, wsfunc)		
			ws.open ()
			
		elif design_spec == skitai.WS_GROUPCHAT:
			# WEBSOCKET_GROUPCHAT
			# /chat?roomid=456, 
			# return (WEBSOCKET_GROUPCHAT, 600)
			# non-threaded websocketserver
			# can send to all clients of group / specific client
			varnames = varnames [:2]
			param_name = varnames [1]
			gid = http_util.crack_query (query).get (param_name, None)
			try:
				assert gid, "%s value can't find" % param_name
			except:
				self.wasc.logger.trace ("server",  request.uri)
				return request.response.error (500, why = apph.debug and sys.exc_info () or None)
			gid = "%s/%s" % (path, gid)
			
			if not servers.websocket_servers.has_key (gid):
				server = servers.websocket_servers.create (gid, self, request, apph, env, message_encoding)				
				if server is None:
					return request.response.error (503)
				env ["websocket"] = server
				if is_atila: env ["websocket.handler"] = (current_app, wsfunc)
			
			server = servers.websocket_servers.get (gid)							
			ws = specs.WebSocket5 (self, request, server, env, varnames)
			self.channel_config (request, ws, keep_alive)
			server.add_client (ws)
		
		request.channel.die_with (ws, "websocket spec.%d" % design_spec)		
		
	def channel_config (self, request, ws, keep_alive):
		request.response.done (upgrade_to =  (ws, 2))
		request.channel.set_socket_timeout (keep_alive)
		


from . import wsgi_handler
import skitai
from aquests.protocols.http import request as http_request
from aquests.protocols.http import request_handler as http_request_handler
from .proxy import POST_MAX_SIZE, UPLOAD_MAX_SIZE
from .proxy.collector import Collector
from .proxy.tunnel import TunnelHandler
from .proxy.response import ProxyResponse
from .proxy import PROXY_TUNNEL_KEEP_ALIVE, PROXY_KEEP_ALIVE

class proxy_request_handler (http_request_handler.RequestHandler):
	def __init__ (self, asyncon, request, callback, client_request, collector, connection = "keep-alive"):
		http_request_handler.RequestHandler.__init__ (self, asyncon, request, callback, connection)
		self.collector = collector
		self.client_request = client_request
		self.is_tunnel = False
		self.should_http2 = False
		self.new_handler = None
		
	def add_reply_headers (self):
		for line in self.response.get_header_lines ():
			try: 
				k, v = line.split (": ", 1)
			except ValueError:
				k, v = line.split (":", 1)								
			ll = k.lower ()
			if ll in ("expires", "date", "connection", "keep-alive", "content-length", "transfer-encoding", "content-encoding", "age", "vary"):
				continue
			if ll == "server":
				self.client_request.response.update ("Server", v.strip () + ", " + skitai.NAME)
				continue
			self.client_request.response [k] = v.strip ()			
	
	ESTABLISHED = b" 200 Connection Established\r\nServer: " + skitai.NAME.encode ("utf8")
	def has_been_connected (self):
		if self.request.method == "connect":
			self.buffer =	b"HTTP/" + self.client_request.version.encode ("utf8") + self.ESTABLISHED
			self.found_terminator ()			
		elif self.should_http2:
			self.switch_to_http2 ()
		
	def will_open_tunneling (self):
		return self.response.code == 200 and self.response.msg.lower () == "connection established"
	
	def connection_closed (self, why, msg):
		# asyncon disconnected
		if self.client_request.channel and self.response:			
			# http 1.0
			self.response.done ()
			#return self.client_request.channel.close_when_done ()
			return
		# abort channel suddenly	
		return http_request_handler.RequestHandler.connection_closed (self, why, msg)
					
	def close_case (self):		
		# unbind readable/writable methods
		if self.asyncon:
			self.asyncon.handler = self.new_handler
			
		if self.callback:			
			self.callback (self)
	
	def found_end_of_body (self):		
		if self.response:
			self.response.done ()
		if self.will_be_close ():
			self.asyncon.disconnect ()
		self.close_case_with_end_tran ()
	
	def create_tunnel (self):
		self.asyncon.established = True		
		self.new_handler = TunnelHandler (self.asyncon, self.request, self.client_request.channel)
		self.client_request.response ("200 Connection Established")
		if self.client_request.version == "1.0":
			self.client_request.version = "1.1"
		self.client_request.response.done (
			upgrade_to = (self.new_handler.asyntunnel, None)			
		)
			
	def create_response (self):
		if not self.client_request.channel: return

		buffer, self.buffer = self.buffer, b""
		accept_gzip = ""
		accept_encoding = self.client_request.get_header ("Accept-Encoding")
		if accept_encoding and accept_encoding.find ("gzip") != -1:
			accept_gzip = "gzip"

		try:
			self.response = ProxyResponse (self.request, buffer.decode ("utf8"), accept_gzip, self.client_request, self.asyncon)
			if self.handle_response_code ():				
				return
		except http_request_handler.ProtocolSwitchError:
			return self.asyncon.handle_close (716)			
		except:
			self.log ("response header error: `%s`" % repr (buffer [:80]), "error")
			return self.asyncon.handle_close (715)
			
		if self.will_open_tunneling ():
			self.create_tunnel ()
			self.close_case ()			
			return
		
		self.client_request.response.start (self.response.code, self.response.msg)
		self.add_reply_headers ()
		
		if self.response.is_gzip_compressed ():
			self.client_request.response ["Content-Encoding"] = "gzip"
				
		if self.response.body_expected ():
			self.client_request.response.push (self.response)
			self.client_request.response.die_with (self.response)
			self.client_request.response.set_streaming ()
		self.client_request.response.done ()
	
	def handle_request (self):
		if not self.client_request.channel: return
		
		self.buffer, self.response = b"", None
		self.asyncon.set_terminator (b"\r\n\r\n")		
		if self.client_request.get_header ('content-type', '').startswith ('application/grpc'):
			self.should_http2 = True
		if self.request.method != "connect" and not self.should_http2:
			self.asyncon.push (self.get_request_header ())
			payload = self.get_request_payload ()
			if payload:
				# don't init_send cause of producer has no data yet
				self.asyncon.push_with_producer (payload, init_send = False)
		self.asyncon.begin_tran (self)
	
	def get_request_payload (self):
		if self.collector:
			self.collector.reuse_cache ()
			if not self.collector.got_all_data:
				self.client_request.set_streaming ()				
		return self.collector
		
	def get_request_header (self, http_version = "1.1"):
		hc = {}	
		address = self.request.get_address ()
		if address [1] in (80, 443):
			hc ["Host"] = "%s" % address [0]
		else:
			hc ["Host"] = "%s:%d" % address			
		hc ["Connection"] = self.connection
		hc ["Accept-Encoding"] = "gzip"
		
		method = self.request.get_method ()			
		additional_headers = self.client_request.get_headers ()
		
		if additional_headers:
			for line in additional_headers:
				try:
					k, v = line.split (": ", 1)
				except ValueError:
					k, v = line.split (":", 1)	
				ll = k.lower ()
				if ll in ("connection", "keep-alive", "accept-encoding", "host"):
					continue
				hc [k] = v
				
		hc ["X-Forwarded-For"] = "%s" % self.client_request.get_remote_addr ()
		hc ["X-Proxy-Agent"] = skitai.NAME
	
		req = "%s %s HTTP/%s\r\n%s\r\n\r\n" % (
			method,
			self.request.path,
			http_version,
			"\r\n".join (["%s: %s" % x for x in hc.items ()])
		)
		
		#print ("####### SKITAI => SERVER ##########################")
		#print (req)
		#print ("---------------------------------")
		return req.encode ("utf8")

			
class Handler (wsgi_handler.Handler):
	def __init__ (self, wasc, clusters, cachefs = None, unsecure_https = False):
		self.wasc = wasc
		self.clusters = clusters
		self.cachefs = cachefs		
		self.unsecure_https = unsecure_https
		self.q = {}
				
	def match (self, request):
		uri = request.uri.lower ()
		if uri.startswith ("http://") or uri.startswith ("ws://") or uri.startswith ("wss://"):
			return 1
		if self.unsecure_https and uri.startswith ("https://"):
			return 1
		if request.command == "connect":
			return 1
		return 0
		
	def handle_request (self, request):
		collector = None		
		if request.command in ('post', 'put'):
			ct = request.get_header ("content-type")
			if not ct: ct = ""
			current_post_max_size = ct.startswith ("multipart/form-data") and UPLOAD_MAX_SIZE or POST_MAX_SIZE
			collector = self.make_collector (Collector, request, current_post_max_size)			
			if collector:
				request.collector = collector				
				collector.start_collect ()
			else:
				return # error was already called in make_collector ()							
		self.continue_request(request, collector)
					
	def continue_request (self, request, collector, asyncon = None):
		if request.command != "connect" and self.has_valid_cache (request, collector is not None):
			return
		
		if asyncon is None:		
			if request.command == "connect":
				asyncon_key = "tunnel://" + request.uri + "/"
				keep_alive = PROXY_TUNNEL_KEEP_ALIVE
			else:
				asyncon_key = request.uri					
				keep_alive = PROXY_KEEP_ALIVE
			asyncon = self.clusters ["__socketpool__"].get (asyncon_key)				
			asyncon.set_keep_alive (keep_alive)
			request.channel.set_socket_timeout (keep_alive)
			if request.command == "connect":
				# refix zombie_timeout becasue it will be not recalcuated
				asyncon.set_timeout (keep_alive)
				
		try:				
			req = http_request.HTTPRequest (request.uri, request.command, collector is not None, logger = self.wasc.logger.get ("server"))				
			r = proxy_request_handler (asyncon, req, self.callback, request, collector)			
			if collector:
				collector.asyncon = asyncon
			r.handle_request ()
				
		except:
			asyncon.set_active (False)
			self.wasc.logger.trace ("server")			
			request.response.error (500, "Proxy Failed")
			
	def has_valid_cache (self, request, has_data):
		if not self.cachefs:
			request.response ["X-Cache-Lookup"] = "NOCACHE"
			return False
		
		if has_data:
			# have collector, cna't cache
			request.response ["X-Cache-Lookup"] = "PASSED"
			return False
				
		cachable = self.cachefs.is_cachable (
			request.get_header ("cache-control"),
			request.get_header ("cookie") is not None, 
			request.get_header ("authorization") is not None, 
			request.get_header ("pragma")			
		)		
		if not cachable:
			request.response ["X-Cache-Lookup"] = "PASSED"
			return False
		
		try:
			accept_encoding = request.get_header ("accept-encoding")
			hit, compressed, max_age, content_type, content = self.cachefs.get (request.uri, accept_encoding and accept_encoding.find ("gzip") != -1)
					
		except:
			self.wasc.logger.trace ("server")	
			return False
		
		else:
			if hit is None:
				request.response ["X-Cache-Lookup"] = "MISS"
				return False
				
		request.response ["X-Cache-Lookup"] = hit == 1 and "MEM_HIT" or "HIT"
		if content_type:
			request.response ["Content-Type"] = content_type					
		request.response ["Cache-Control"] = "max-age=%d" % max_age
		if compressed:
			request.response ["Content-Encoding"] = "gzip"
		request.response.push (content)
		request.response.done ()
		return True
	
	def save_cache (self, request, handler):
		try:			
			if self.cachefs and not handler.request.payload and handler.response.max_age:
				self.cachefs.save (
					request.uri, 
					handler.response.get_header ("content-type"), handler.response.content, 
					handler.response.max_age, request.response ["Content-Encoding"] == "gzip"
				)
				
		except:
			self.wasc.logger.trace ("server")
	
	def dealloc (self, request, handler):
		handler.callback = None				
		handler.response = None
		handler.collector = None
		request.collector = None
		request.producer = None
		request.response = None # break back ref.		
		del handler
			
	def callback (self, handler):
		response, request = handler.response, handler.client_request

		if request.channel:
			if response.code >= 700:
				request.response.error (506, response.msg)
			
			else:
				try:
					self.save_cache (request, handler)
				except:
					self.wasc.logger.trace ("server")
		
		self.dealloc (request, handler)
		


# was.py
# synchronouse WAS rquest service for app tesing

from .. import wsgiappservice
from ..rpc import cluster_dist_call
from ..dbi import cluster_dist_call as dcluster_dist_call 
from skitai import DB_SQLITE3, DB_PGSQL, DB_REDIS, DB_MONGODB
from rs4 import webtest
import random
from urllib.parse import urlparse, urlunparse       
from ..wastuff import modelkeys
 
class Result:
    def __init__ (self, status, response = None):
        self.status = status
        self.__response = response
    
    def __getattr__ (self, attr):
        return getattr (self.__response, attr)
            
    @property
    def data (self):
        if hasattr (self.__response, "status_code"):                
            ct = self.__response.headers.get ("content-type", "")
            if ct:
                if ct.startswith ("application/json"):
                    return self.__response.json ()
            return self.__response.text
        return self.__response    
    
class ProtoCall (cluster_dist_call.ClusterDistCall):
    def __init__ (self, cluster, *args, **kargs):
        self.cluster = cluster
        self.result = None   
        self.handle_request (*args, **kargs)       
            
    def  handle_request (self, uri, params = None, reqtype="rpc", headers = None, auth = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, timeout = 10, caller = None):
        if self.cluster:        
            endpoints = self.cluster.get_endpoints ()
            endpoint = random.choice (endpoints)
        else:
            parts = urlparse (uri)
            endpoint = "{}://{}".format (parts [0], parts [1])
            uri = urlunparse (("", "") + parts [2:]) 
                     
        with webtest.Target (endpoint) as cli:
            req_func = getattr (cli, reqtype)
            try:
                resp = req_func (uri, headers = headers, auth = auth)                
            except:
                self.result = Result (1)
            else:
                self.result = Result (3, resp)        
        callback and callback (self.result)
    
    def wait (self):
        pass
    
    def _or_throw (self, func, status, timeout, cache):
        if self.result.status != 3:        
            raise exceptions.HTTPError (status, sys.exc_info ())
        return self.result
            
    def getwait (self):
        return self.result
    
    def getswait (self):
        return cluster_dist_call.Results ([self.result])
        

class DBCall (ProtoCall):
    def __init__ (self, cluster, *args, **kargs):
        self.cluster = cluster
        self.result = None
        self.args, self.kargs = args, kargs
    
    def _build_request (self, method, param):
        self.handle_request (method, param, *self.args, **self.kargs)
        
    def  handle_request (self, method, param, server = None, dbname = None, auth = None, dbtype = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, timeout = 10, caller = None):
        from ..dbi import cluster_manager
        
        assert dbtype is None, "please, alias {}".format (server)
        if self.cluster:
            conns = self.cluster.get_endpoints ()
        else:            
            conns = cluster_manager.make_endpoints (dbtype, [server, dbname, auth])            
        conn = random.choice (conns)
        try:
            if self.cluster.dbtype in (DB_SQLITE3, DB_PGSQL):        
                cur = conn.cursor ()
                getattr (cur, method) (*param)
                resp = cur.fetchall ()
            else:
                resp = getattr (conn, method) (*param)
        except:
            self.result = Result (1)
        else:
            self.result = Result (3, resp)                    
        for conn in conns:
            conn.close ()
        callback and callback (self.result)

class ModelKeys (modelkeys.ModelKeys):
    def __init__ (self):
        self._arr = [0] * 1024        
        self._d = {}
            
    def set (self, k, v, ignore_nokey = False):
        if k not in self._d:
            self._d [k] = len (self._d)
        self._arr [self._d [k]] = v
        
class WAS (wsgiappservice.WAS):
    numthreads = 1 
    _luwatcher = ModelKeys ()
    
    def _create_rest_call (self, cluster, *args, **kargs):
        return ProtoCall (cluster, *args, **kargs)
    
    def _create_dbo (self, cluster, *args, **kargs):
        return dcluster_dist_call.Proxy (DBCall, cluster, *args, **kargs)
    


import re, sys, os
try:
    from urllib.parse import unquote, unquote_plus
except ImportError:
    from urlparse import unquote
    from urllib import unquote_plus    
import json
from hashlib import md5
import random

####################################################################################
# Utitlities
####################################################################################
REQUEST = re.compile ('([^ ]+) ([^ ]+)(( HTTP/([0-9.]+))$|$)')
CONNECTION = re.compile ('Connection: (.*)', re.IGNORECASE)

def crack_query (r):
    if type (r) is bytes:
        r = r.decode ("utf8")

    if not r: return {}
    if r[0]=='?': r=r[1:]    
    arg={}
    q = [x.split('=', 1) for x in r.split('&')]
    
    for each in q:
        k = unquote_plus (each[0])
        try: 
            t, k = k.split (":", 1)
        except ValueError:
            t = "str"
            
        if len (each) == 2:        
            v = unquote_plus(each[1])
            if t == "str":
                pass
            elif t == "int":
                v = int (v)
            elif t == "float":
                v = float (v)
            elif t == "list":
                v = v.split (",")
            elif t == "bit":
                v = v.lower () in ("1", "true", "yes")
            elif t == "json":
                v = json.loads (v)
                
        else:
            v = ""            
            if t == "str":
                pass
            elif t == "int":
                v = 0
            elif t == "float":
                v = 0.0
            elif t == "list":
                v = []
            elif t == "bit":
                v = False
            elif t == "json":
                v = {}
                
        if k in arg:
            if type (arg [k]) is not type ([]):
                arg[k] = [arg[k]]
            arg[k].append (v)
            
        else:
            arg[k] = v
            
    return arg
    
def crack_request (r):
    m = REQUEST.match (r)
    if m.end() == len(r):        
        uri=m.group(2)        
        if m.group(3):
            version = m.group(5)
        else:
            version = None    
        return m.group(1).lower(), uri, version
    else:
        return None, None, None

def join_headers (headers):
    r = []
    for i in range(len(headers)):
        if headers[i][0] in ' \t':    
            r[-1] = r[-1] + headers[i][1:]
        else:
            r.append (headers[i])
    return r

def get_header (head_reg, lines, group=1):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m.group (group)
    return ''

def get_header_match (head_reg, lines):
    for line in lines:
        m = head_reg.match (line)
        if m and m.end() == len(line):
            return m
    return ''        

def get_extension (path):
    dirsep = path.rfind ('/')
    dotsep = path.rfind ('.')
    if dotsep > dirsep:
        return path[dotsep+1:]
    else:
        return ''

ALNUM = '0123456789abcdefghijklmnopqrstuvwxyz'
def md5uniqid (length = 13):    
    global ALNUM
    _id = ''
    for i in range (0, length):
        _id += random.choice(ALNUM)
    return md5 (_id.encode ("utf8")).hexdigest ()[length:]
    